"""Top-level package for eddymotion."""
from eddymotion._version import __version__

__packagename__ = "eddymotion"
__copyright__ = "Copyright 2021, The eddymotion developers"
__url__ = "https://github.com/nipreps/EddyMotionCorrection"

DOWNLOAD_URL = (
    f"https://github.com/nipreps/{__packagename__}/archive/{__version__}.tar.gz"
)
